/*
 * author: bernhard aichinger
 * matnr:  9825670
 *
 *
 * a program class: implements the termination detection algorithm by dijsktra
 *
 */

import daj.*;
import java.util.Random;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.NoSuchElementException;


class Prog extends Program {

    private static final char regionREM = 'R';
    private static final char regionTRY = 'T';
    private static final char regionCRIT = 'C';
    private static final char regionEXIT = 'E';


    private int rank;
    private int size;

    // state variables according to the automaton
    public char region;
    public int clock;
    public LinkedList bcast;
    public LinkedList[] send;
    public LinkedList[] history;


    public Prog(int rank, int size) { 
        this.rank = rank;
        this.size = size;

        // state variables
        region = regionREM;
        clock = 0;
        bcast = new LinkedList();
        send = new LinkedList[size];
        history = new LinkedList[size];

        for (int i = 0; i < size; i++) {
            send[i] = new LinkedList();
            history[i] = new LinkedList();
        }
    }


    private void dumpHistory() {
        System.out.println("-------------------------------------");
        System.out.println("history of " + rank);
        for (int i = 0; i < history.length; i++) {
            System.out.println("node " + i);
            Iterator iter = history[i].iterator();
            while (iter.hasNext())
                System.out.println(((Msg)iter.next()).getText());
        }
        System.out.println("-------------------------------------");
    }


    /*
     * execution of program
     */
    public void main() { 

        Random rnd = new Random();

        while (true) {
            switch(rnd.nextInt(7)) {
            case 0:      // try
                if (region == regionEXIT || region == regionREM)
                transitionTRY();
                // transitionBCAST();
                break;
            case 1:      // bcast
                transitionBCAST();
                break;
            case 2:      // receive
                transitionRECEIVE();
                break;
            case 3:      // send
                transitionSEND();
                break;
            case 4:      // crit
                if (region == regionTRY)
                transitionCRIT();
                break;
            case 5:      // exit
                if (region == regionCRIT || region == regionTRY)
                transitionEXIT();
                break;
            case 6:      // rem
                if (region == regionEXIT)
                transitionREM();
                break;
            default:
                break;
            }
        }
    }


    private void transitionTRY() {
        clock++;
        region = regionTRY;
        dumpState();
        bcast.addLast(new Msg(Msg.TRY, clock, rank));
    }


    private void transitionBCAST() {
        try {
            Msg msg = (Msg)bcast.getFirst();
            // int c = msg.getClock();

            // if (c == clock + 1) {
            // clock = c;

            clock++;
            bcast.removeFirst();
            for (int i = 0; i < out().getSize(); i++)
                out(i).send(msg);
            
            //System.out.println(rank + " (" + clock + "): bcast " + msg.getText());

            // }

        }
        catch (NoSuchElementException e) {
        }
    }


    private void transitionRECEIVE() {
        int sender = in().select(1);
        while (sender != -1) {
            Msg msg = (Msg)in(sender).receive();
            if (msg != null) {
                int c = msg.getClock();
                int m = msg.getType();
                
                clock = (clock > c ? clock : c) + 1;
                history[sender].addLast(msg);
                if (m == Msg.TRY && sender != rank)
                    send[sender].addLast(new Msg(Msg.ACK, clock, rank));

                //System.out.println(rank + " (" + clock + "): receive " + msg.getText() + " from " + sender);
            }
            sender = in().select(1);
        }
    }


    private void transitionSEND() {
        for (int j = 0; j < out().getSize(); j++) {
            try {
                Msg msg = (Msg)send[j].getFirst();
                // int c = msg.getClock();
                
                // if (c == clock + 1) {
                // clock = c;
                
                clock++;
                send[j].removeFirst();
                out(j).send(msg);
                
                //System.out.println(rank + ": send " + msg.getText() + " to " + j);
            }
            catch (NoSuchElementException e) {
            }
        }
    }



    private void transitionCRIT() {
        if (region == regionTRY) {

            boolean
                foundTry = false,
                foundOther = true,
                foundAck = false;

            //dumpHistory();

            Iterator iterTry = history[rank].iterator();
            while (iterTry.hasNext() &&
                   !(foundTry && !foundOther && foundAck)) {

                /*
                 * search TRY
                 * there must not be an EXIT which is newer than TRY
                 */
                foundTry = false;
                boolean foundExit = false;
                Msg msgTry = (Msg)iterTry.next();
                if (msgTry.getType() == Msg.TRY) {
                    //System.out.println("- my try: " + msgTry.getText());
                    foundTry = true;
                    Iterator iterExit = history[rank].iterator();
                    while (iterExit.hasNext()) {
                        Msg msgExit = (Msg)iterExit.next();
                        if (msgExit.getType() == Msg.EXIT) {
                            if (msgExit.getClock() > msgTry.getClock()) {
                                //System.out.println("- my exit " + msgExit.getText());
                                foundExit = true;
                            }
                        }
                    }
                }
                foundTry = foundTry && !foundExit;

                // if (foundTry) System.out.println(rank + " foundTry");


                /*
                 * search TRY/EXIT pairs of other nodes
                 * if a TRY of an other node exist, then there must be a EXIT too
                 */
                foundOther = false;
                for (int j = 0; j < history.length; j++) {
                    if (j != rank) {
                        Iterator iterTry2 = history[j].iterator();
                        while (iterTry2.hasNext()) {
                            boolean
                                foundOtherTry = false,
                                foundOtherExit = false;
                            Msg msgTry2 = (Msg)iterTry2.next();
                            if (msgTry2.getType() == Msg.TRY &&
                                (msgTry2.getClock() < msgTry.getClock() ||
                                 (msgTry2.getClock() == msgTry.getClock() &&
                                  j < rank))) {
                                //System.out.println("- " + rank + ": other try from " + j + " " + msgTry2.getText());
                                foundOtherTry = true;
                                Iterator iterExit2 = history[j].iterator();
                                while (iterExit2.hasNext()) {
                                    Msg msgExit2 = (Msg)iterExit2.next();
                                    if (msgExit2.getType() == Msg.EXIT) {
                                        if (msgExit2.getClock() > msgTry2.getClock()) {
                                            //System.out.println("- " + rank + ": other exit from " + j + " " + msgExit2.getText());
                                            foundOtherExit = true;
                                        }
                                    }
                                } // while
                            }
                            foundOther = foundOther || (foundOtherTry ^ foundOtherExit);
                        } // while
                    }
                } // for

                // if (!foundOther) System.out.println(rank + " !foundOther");

                /*
                 * search ACK, all nodes must acknowlegde
                 */
                int counter = 0;
                for (int j = 0; j < history.length; j++) {
                    if (j != rank) {
                        Iterator iter = history[j].iterator();
                        while (iter.hasNext()) {
                            Msg msg = (Msg)iter.next();
                            if (msg.getType() == Msg.ACK)
                                if (msgTry.getClock() < msg.getClock() ||
                                    (rank < j && msgTry.getClock() == msg.getClock())) {
                                    //System.out.println("- " + rank + " ack from " + j + " " + msg.getText());
                                    counter++;
                                }
                        }
                    }
                } // for
                foundAck = (counter == (history.length - 1));

                // if (foundAck) System.out.println(rank + " foundAck");

            } // while

            if (foundTry && !foundOther && foundAck) {
                clock++;
                region = regionCRIT;

                dumpState();

                // interrupt();
            }
        }
    }


    private void transitionEXIT() {
        clock++;
        region = regionEXIT;
        dumpState();
        bcast.addLast(new Msg(Msg.EXIT, clock, rank));
    }


    private void transitionREM() {
        if (region == regionEXIT) {
            boolean found = false;


            Iterator iterExit = history[rank].iterator();
            while (iterExit.hasNext()) {
                Msg msgExit = (Msg)iterExit.next();
                if (msgExit.getType() == Msg.EXIT) {
                    Iterator iterTry = history[rank].iterator();
                    while (iterTry.hasNext()) {
                        Msg msgTry = (Msg)iterTry.next();
                        if (msgTry.getType() == Msg.TRY) {
                            if (msgTry.getClock() > msgExit.getClock())
                                found = true;
                        }
                    }
                }
            }
            if (!found) {
                clock++;
                region = regionREM;
                dumpState();
            }
        }
    }



    /*
     * reflects the state of a node
     */
    public String getText() {
        String msgString = "";

        msgString += "region: " + region + "\n";
        msgString += "clock: " + clock;

        return msgString;
    }

    private void dumpState() {
        System.out.println(">> " + rank + " : " + region + " / " + clock);
    }

}
